import json
from abc import ABC, abstractmethod
from typing import Any, Dict, Iterable

from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.reader.base import BaseReader


class JsonReader(BaseReader, ABC):
    """
    Base class for reading json-based files
    """

    def _read(self, file: Iterable[str]) -> GeoObjectFile:
        json_file = "\n".join(file)
        loaded: Dict[Any, Any] = json.loads(json_file)

        return self.read_json(loaded)

    @abstractmethod
    def read_json(self, json_dict: Dict[Any, Any]) -> GeoObjectFile:
        """
        Read implementation for json based files
        :param json_dict: parsed json representation
        :return: Domain representation of the GeoObjectFile
        """
        return GeoObjectFile()



import re
from abc import ABC, abstractmethod
from io import TextIOWrapper
from typing import Any, Dict, Generator, Iterable, List

from geofiles.domain.face import Face
from geofiles.domain.geo_object_file import GeoObjectFile


class BaseReader(ABC):
    """
    BaseReader implementation for importing geo referenced objects
    """

    def read(self, file: Any) -> GeoObjectFile:
        """
        Reads a given file
        :param file: to be read (may be a string representing the path or an opened file instance)
        :return: Domain representation of the GeoObjectFile
        """
        close = False
        to_read: Any = None
        try:
            if isinstance(file, str):
                to_read = open(file)
                close = True
            elif isinstance(file, TextIOWrapper):
                to_read = file
            else:
                raise Exception(f"Can't handle {file}")

            return self._read(BaseReader._read_file_line(to_read))
        finally:
            if close and to_read is not None:
                to_read.close()

    def read_strings(self, input_strings: List[str]) -> GeoObjectFile:
        """
        Reads a given geometric object
        :param input_strings: strings describing a geometric object
        """
        return self._read(input_strings)

    def read_string(self, input_string: str) -> GeoObjectFile:
        """
        Reads a given geometric object
        :param input_string: string describing a geometric object
        """
        return self._read(BaseReader._split_str(input_string, "\n"))

    @staticmethod
    def _read_file_line(file: TextIOWrapper) -> Generator[str, None, None]:
        """
        Lazy function (generator) to read a file line by line
        :param file: to be read
        """
        while True:
            data = file.readline()
            if not data:
                break
            yield data

    @staticmethod
    def _split_str(string: str, sep: str = "\n") -> Generator[str, None, None]:
        """
        Lazy function (generator) to read a string line by line separated by the given sep symbol
        :param string: to be read
        """
        # warning: does not yet work if sep is a lookahead like `(?=b)`
        if sep == "":
            return (c for c in string)
        return (
            _.group(1) for _ in re.finditer(f"(?:^|{sep})((?:(?!{sep}).)*)", string)
        )

    @abstractmethod
    def _read(self, file: Iterable[str]) -> GeoObjectFile:
        """
        Read implementation
        :param file: target to be read
        :return: Domain representation of the GeoObjectFile
        """
        return GeoObjectFile()

    @staticmethod
    def _filter_faces(
        unique_vertices: bool,
        coordinate: List[Any],
        face_object: Face,
        vertex_list: List[List[Any]],
        vertex_indices: Dict[str, int],
    ) -> None:
        """
        Implementation for filtering non_unique vertices
        :param unique_vertices: Flag if vertices should be filtered
        :param coordinate: Current coordinate
        :param face_object: Current face
        :param vertex_list: List of all resulting vertices
        :param vertex_indices: Dictionary mapping already known vertices (in str representation) with their idx in the list
        :return: None
        """
        if unique_vertices:
            str_rep = "-".join([str(a) for a in coordinate])
            if not vertex_indices.get(str_rep):
                vertex_indices[str_rep] = len(vertex_list) + 1
                vertex_list.append(coordinate)

            vertex_index = vertex_indices[str_rep]
            face_object.indices.append(vertex_index + 1)
        else:
            face_object.indices.append(len(vertex_list))
            vertex_list.append(coordinate)



from abc import ABC
from typing import Any, Dict, List

from geofiles.domain.face import Face
from geofiles.domain.geo_object import GeoObject
from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.reader.base import BaseReader
from geofiles.reader.json_reader import JsonReader


class GeoJsonReader(JsonReader, BaseReader, ABC):
    """
    Reader implementation for GeoJSON files
    Note: That only certain FeatureCollection GeoJSONs with Polygons or MultiPolygons are supported
    """

    def __init__(self):
        """
        unique_vertices: defines that read vertices have to be unique
        """
        self.unique_vertices = False

    def read_json(self, json_dict: Dict[Any, Any]) -> GeoObjectFile:
        if not json_dict.get("type") or json_dict.get("type") != "FeatureCollection":
            raise Exception("Only GeoJSONs with type FeatureCollection are supported")

        result = GeoObjectFile()
        result.crs = "urn:ogc:def:crs:OGC:2:84"
        vertex_list: List[List[Any]] = []
        vertex_indices: Dict[str, int] = dict()

        features = json_dict.get("features")
        if features:
            for feature in features:
                if feature.get("type"):
                    feature_type = feature.get("type")
                    coordinates = feature.get("coordinates")

                    faces = []
                    if feature_type == "MultiPolygon":
                        for coordinate in coordinates:
                            inner_coordinates: List[Any] = []
                            self.get_values_of_most_inner_array(
                                coordinate, inner_coordinates
                            )
                            faces.append(
                                inner_coordinates[: len(inner_coordinates) - 1]
                            )
                    elif feature_type == "Polygon":
                        inner_coordinates = []
                        self.get_values_of_most_inner_array(
                            coordinates, inner_coordinates
                        )
                        faces.append(inner_coordinates[: len(inner_coordinates) - 1])

                    geo_object = GeoObject()
                    result.objects.append(geo_object)
                    for face in faces:
                        face_object = Face()
                        geo_object.faces.append(face_object)
                        for coordinate in face:
                            self._filter_faces(
                                self.unique_vertices,
                                coordinate,
                                face_object,
                                vertex_list,
                                vertex_indices,
                            )
        result.vertices = vertex_list
        return result

    def get_values_of_most_inner_array(
        self, input_list: List[Any], res: List[Any]
    ) -> None:
        """
        iterates the given list and fills the res list with the most inner values
        """
        if len(input_list) > 0:
            elem = input_list[0]
            if isinstance(elem, list):
                for element in input_list:
                    self.get_values_of_most_inner_array(element, res)
            else:
                res.append(input_list)



import json
from abc import ABC, abstractmethod
from io import TextIOWrapper
from typing import Any, Dict

from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.writer.base import BaseWriter


class JsonWriter(BaseWriter, ABC):
    """
    Basic writer implementations for json based file formats
    """

    @abstractmethod
    def create_json(
        self, data: GeoObjectFile, random_seed: Any = None
    ) -> Dict[Any, Any]:
        """
        Creates a json representation of the given GeoObjectFile
        :param data: to be written to json
        :param random_seed: used for id generation
        """
        return dict()

    def write_to_string(
        self, data: GeoObjectFile, write_binary: bool = False, random_seed: Any = None
    ) -> str:
        return json.dumps(self.create_json(data, random_seed))

    def _write(
        self,
        file: TextIOWrapper,
        data: GeoObjectFile,
        write_binary: bool,
        random_seed: Any,
    ) -> None:
        json.dump(self.create_json(data, random_seed), file)



from abc import ABC
from typing import Any, Dict, List

from geofiles.domain.face import Face
from geofiles.domain.file_version import CityJsonVersion
from geofiles.domain.geo_object import GeoObject
from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.reader.base import BaseReader
from geofiles.reader.json_reader import JsonReader


class CityJsonReader(JsonReader, BaseReader, ABC):
    """
    Reader implementation for CityJSON files
    Note: That there will be semantic loss if reading a CityJSON file (object classes, etc.)
    """

    def __init__(
        self,
        version: CityJsonVersion = CityJsonVersion.V1_1,
        ignore_mandatory_transform: bool = False,
    ):
        """
        :param version: The CityJSON version to be used
        :param ignore_mandatory_transform: Starting with CityJSON v1.1 the Transform information is mandatory. Ignore this, when reading the file.
        """
        self.version = version
        self.ignore_mandatory_transform = ignore_mandatory_transform

    def read_json(self, json_dict: Dict[Any, Any]) -> GeoObjectFile:
        result = GeoObjectFile()

        if not json_dict.get("metadata"):
            raise Exception(
                "No metadata defined (at least reference system is required)"
            )

        metadata = json_dict["metadata"]

        transform = json_dict.get("transform")
        if (
            self.version != CityJsonVersion.V1_0
            and transform is None
            and not self.ignore_mandatory_transform
        ):
            raise Exception("Transform information is mandatory in CityJSON >= 1.1")

        if transform is not None:
            translate = transform["translate"]
            result.translation = translate
            scale = transform["scale"]
            result.scaling = scale

        if metadata.get("referenceSystem"):
            result.crs = metadata.get("referenceSystem")
        else:
            raise Exception("Unknown reference system in input file.")

        if metadata.get("geographicalExtent"):
            extents = metadata["geographicalExtent"]
            result.min_extent = extents[:3]
            result.max_extent = extents[3:]

        if json_dict.get("vertices"):
            result.vertices = json_dict["vertices"]
        else:
            raise Exception("Undefined vertices in input file.")

        if json_dict.get("CityObjects"):
            city_objects = json_dict["CityObjects"]
            for city_object_name, city_object in city_objects.items():
                geo_object = GeoObject()
                result.objects.append(geo_object)
                geo_object.name = city_object_name
                geometry = city_object.get("geometry")
                if geometry:
                    for geometry_object in geometry:
                        boundaries = geometry_object.get("boundaries")
                        if boundaries:
                            faces: List[Face] = []
                            self.get_values_of_most_inner_array(boundaries, faces)
                            geo_object.faces = faces
        else:
            raise Exception("No city objects defined")

        return result

    def get_values_of_most_inner_array(
        self, input_list: List[Any], res: List[Any]
    ) -> None:
        """
        iterates the given list and fills the res list with the most inner values
        :param input_list: to be iterated
        :param res: Result list containing the most inner list elements
        """
        if len(input_list) > 0:
            elem = input_list[0]
            if isinstance(elem, list):
                for element in input_list:
                    self.get_values_of_most_inner_array(element, res)
            else:
                face = Face()
                for e in input_list:
                    face.indices.append(e + 1)
                res.append(face)



from abc import ABC
from typing import Iterable

from geofiles.domain.face import Face
from geofiles.domain.geo_object import GeoObject
from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.reader.base import BaseReader


class GeoOffReader(BaseReader, ABC):
    """
    Reader implementaiton for geo-referenced .off files (.geooff)
    """

    def _read(self, file: Iterable[str]) -> GeoObjectFile:
        res = GeoObjectFile()
        obj = GeoObject()
        res.objects.append(obj)
        num_of_vertices = 0
        search_for_vertices = False
        cnt = 0

        headerdefinition = []
        found_headerdefinition = False

        for line in file:
            trimmed = line.strip()
            if not trimmed:
                continue
            trimmed = " ".join(trimmed.split())

            if not found_headerdefinition:
                # We have not found a header definition yet, so search for a GeoOFF or OFF header
                idx = trimmed.find("GeoOFF")
                if idx >= 0:
                    # found a GeoOFF header, so will look for the crs definition + further options
                    headerdefinition.append("crs")
                    postfix = trimmed[idx + len("GeoOFF") :]
                    found_headerdefinition = True
                    for e in postfix:
                        headerdefinition.append(e)
                    headerdefinition.append("header")
                elif trimmed.endswith("OFF"):
                    # found a classic OFF file
                    found_headerdefinition = True
                    headerdefinition.append("header")
            elif search_for_vertices:
                # We already looking for vertices and faces, since we know the number of vertices, just count to know when the faces start
                splits = trimmed.split(" ")
                if cnt < num_of_vertices:
                    coordinates = [float(a) for a in splits]
                    res.vertices.append(
                        coordinates[:3]
                    )  # we are only interested in the coordinates and not in optional additional information
                    cnt += 1
                else:
                    face = Face()
                    face.indices = [int(a) + 1 for a in splits[1:]]
                    obj.faces.append(face)
            elif found_headerdefinition:
                element = headerdefinition.pop(0)
                splits = trimmed.split(" ")

                if element == "crs":
                    res.crs = splits[0]
                elif element == "header":
                    num_of_vertices = int(splits[0])
                    search_for_vertices = True
                elif element == "o":
                    res.origin = [float(a) for a in splits]
                elif element == "e":
                    extents = [float(a) for a in splits]
                    res.min_extent = extents[:3]
                    res.max_extent = extents[3:]
                elif element == "s":
                    res.scaling = [float(a) for a in splits]
                elif element == "t":
                    res.translation = [float(a) for a in splits]
                elif element == "r":
                    res.rotation = [float(a) for a in splits]
        return res



import math
from abc import ABC
from io import TextIOWrapper
from typing import Any, Dict, List

from geofiles.conversion.calculation import convert_obj_index
from geofiles.conversion.static import get_wgs_84
from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.writer.base import BaseWriter


class GeoVrmlWriter(BaseWriter, ABC):
    """
    Writer implementation for creating Geo-Referenced VRML geometry files (.wrl)
    """

    def _write(
        self,
        file: TextIOWrapper,
        data: GeoObjectFile,
        write_binary: bool,
        random_seed: Any,
    ) -> None:
        """
        Write implementation
        :param file: target to be written
        :param data: content to be written
        :param write_binary: flag if file is a binary file
        :return:
        """
        if data.crs != get_wgs_84():
            raise Exception(
                "This writer implementation requires WGS:84 coordinate system"
            )

        if write_binary:
            self._write_to_file(file, "#VRML V2.0 ascii", write_binary, True)
        else:
            self._write_to_file(file, "#VRML V2.0 utf8", write_binary, True)

        self._write_to_file(file, "EXTERNPROTO GeoCoordinate [", write_binary, True)
        self._write_to_file(file, " field SFNode geoOrigin # NULL", write_binary, True)
        self._write_to_file(
            file, ' field MFString geoSystem # [ "GDC" ]', write_binary, True
        )
        self._write_to_file(file, " field MFString point # []", write_binary, True)
        self._write_to_file(file, " field MFString point # []", write_binary, True)
        self._write_to_file(
            file,
            '] [ "urn:web3d:geovrml:1.0/protos/GeoCoordinate.wrl"',
            write_binary,
            True,
        )
        self._write_to_file(
            file,
            ' "file:///C|/Program%20Files/GeoVRML/1.0/protos/GeoCoordinate.wrl"',
            write_binary,
            True,
        )
        self._write_to_file(
            file,
            ' "http://www.geovrml.org/1.0/protos/GeoCoordinate.wrl" ]',
            write_binary,
            True,
        )

        if data.is_origin_based() and data.origin is not None:
            self._write_to_file(file, "EXTERNPROTO GeoOrigin [", write_binary, True)
            self._write_to_file(
                file,
                '            exposedField  MFString  geoSystem    # [ "GDC" ]',
                write_binary,
                True,
            )
            self._write_to_file(
                file,
                '            exposedField  SFString  geoCoords    # ""',
                write_binary,
                True,
            )
            self._write_to_file(
                file,
                '          ] [ "urn:web3d:geovrml:1.0/protos/GeoOrigin.wrl"',
                write_binary,
                True,
            )
            self._write_to_file(
                file,
                '	      "file:///C|/Program%20Files/GeoVRML/1.0/protos/GeoOrigin.wrl"',
                write_binary,
                True,
            )
            self._write_to_file(
                file,
                '	      "http://www.geovrml.org/1.0/protos/GeoOrigin.wrl" ]',
                write_binary,
                True,
            )

            origin = " ".join([str(a) for a in data.origin])
            self._write_to_file(file, "DEF ORIGIN GeoOrigin {", write_binary, True)
            self._write_to_file(file, '   geoSystem ["GDC"]', write_binary, True)
            self._write_to_file(file, f'   geoCoords "{origin}"', write_binary, True)
            self._write_to_file(file, "}", write_binary, True)

        num_of_vertices = len(data.vertices)
        num_of_objects = len(data.objects)
        i = 0

        for obj in data.objects:
            local_vertices: List[Any] = []
            vertex_mapping: Dict[int, int] = dict()
            for face in obj.faces:
                for idx in face.indices:
                    if vertex_mapping.get(idx) is None:
                        vertex_mapping[idx] = len(local_vertices)
                        local_vertices.append(data.get_vertex(idx))

            self._write_to_file(
                file, f"DEF OBJECT-{i} Shape " + "{", write_binary, True
            )

            self._write_to_file(
                file, "   geometry IndexedFaceSet {", write_binary, True
            )
            self._write_to_file(file, "      coord GeoCoordinate {", write_binary, True)
            self._write_to_file(
                file, '         geoSystem [ "GDC" ]', write_binary, True
            )
            if data.is_origin_based():
                self._write_to_file(file, "         geoOrigin", write_binary, True)
                self._write_to_file(file, "           USE ORIGIN ", write_binary, True)

            self._write_to_file(file, "         point [", write_binary, True)
            for vertex in local_vertices:
                self._write_to_file(file, "            ", write_binary, False)
                for coord in vertex:
                    self._write_to_file(file, f"{coord} ", write_binary, False)
                self._write_to_file(file, "", write_binary, True)
            self._write_to_file(file, "         ]", write_binary, True)
            self._write_to_file(file, "      }", write_binary, True)
            self._write_to_file(file, "      coordIndex [", write_binary, False)

            num_of_faces = len(obj.faces)
            j = 0
            for face in obj.faces:
                for idx in face.indices:
                    self._write_to_file(
                        file,
                        f"{convert_obj_index(int(idx), num_of_vertices)} ",
                        write_binary,
                        False,
                    )
                self._write_to_file(file, "-1", write_binary, False)
                if j < num_of_faces - 1:
                    self._write_to_file(file, ", ", write_binary, False)
                j += 1

            self._write_to_file(file, "]", write_binary, True)
            self._write_to_file(file, "   }", write_binary, True)
            self._write_to_file(file, "}", write_binary, True)
            i += 1

        if data.scaling is not None:
            self._print_transform(
                file, write_binary, "scale", data.scaling, True, num_of_objects
            )

        if data.rotation is not None:
            self._print_rotation(
                file, write_binary, data.rotation, True, num_of_objects
            )

        if data.translation is not None:
            self._print_transform(
                file,
                write_binary,
                "translation",
                data.translation,
                True,
                num_of_objects,
            )

        i = 0
        for geoobj in data.objects:
            if geoobj.rotation is not None and geoobj.contains_rotation():
                self._print_rotation(
                    file, write_binary, geoobj.rotation, False, object_id=i
                )
            if geoobj.scaling is not None and geoobj.contains_scaling():
                self._print_transform(
                    file, write_binary, "scale", geoobj.scaling, False, object_id=i
                )
            if geoobj.translation is not None and geoobj.contains_translation():
                self._print_transform(
                    file,
                    write_binary,
                    "translation",
                    geoobj.translation,
                    False,
                    object_id=i,
                )

            i += 1

    def _print_rotation(
        self,
        file: TextIOWrapper,
        write_binary: bool,
        to_write: List[Any],
        print_children: bool = True,
        num_of_objects: int = 0,
        object_id: int = 0,
    ) -> None:
        """
        Prints rotation information to the VRML file
        :param file: target to be written
        :param write_binary: flag if file is a binary file
        :param to_write: values to write
        :param print_children: write generic child names
        :param num_of_objects: number of childs if print_children is True
        :param object_id: id of the object if print_children is False
        """
        for idx, rotation in enumerate(to_write):
            rotation = math.radians(rotation)
            if idx == 0:
                axis = [1, 0, 0]
            elif idx == 1:
                axis = [0, 1, 0]
            else:
                axis = [0, 0, 1]

            if not math.isclose(rotation, 0):
                axis.append(rotation)
                self._print_transform(
                    file,
                    write_binary,
                    "rotation",
                    axis,
                    print_children,
                    num_of_objects,
                    object_id,
                )

    def _print_transform(
        self,
        file: TextIOWrapper,
        write_binary: bool,
        prefix: str,
        to_write: List[Any],
        print_children: bool = True,
        num_of_objects: int = 0,
        object_id: int = 0,
    ) -> None:
        """
        Prints transformation information to the VRML file
        :param file: target to be written
        :param write_binary: flag if file is a binary file
        :param prefix: VRML prefix used for the transformation (rotation, translation or scale)
        :param to_write: values to write
        :param print_children: write generic child names
        :param num_of_objects: number of childs if print_children is True
        :param object_id: id of the object if print_children is False
        """
        self._write_to_file(file, "Transform { ", write_binary, True)
        values = " ".join([str(a) for a in to_write])
        self._write_to_file(file, f"   {prefix} {values}", write_binary, True)
        if print_children:
            self._print_children(num_of_objects, file, write_binary)
        else:
            self._write_to_file(file, "   children [", write_binary, True)
            self._write_to_file(
                file, f"      USE OBJECT-{object_id}", write_binary, True
            )
            self._write_to_file(file, "   ]", write_binary, True)
        self._write_to_file(file, "}", write_binary, True)

    def get_file_type(self) -> str:
        """
        :return: the supported file type of this writer
        """
        return ".wrl"

    def _print_children(
        self, num_of_objects: int, file: TextIOWrapper, write_binary: bool
    ) -> None:
        """
        Prints the children definition for a transform element
        :param num_of_objects: number of objects
        :param file: target to be written
        :param write_binary: flag if file is a binary file
        :return: None
        """
        self._write_to_file(file, "   children [", write_binary, True)
        for i in range(0, num_of_objects):
            self._write_to_file(file, f"      USE OBJECT-{i}", write_binary, True)
        self._write_to_file(file, "   ]", write_binary, True)

    def supports_origin_base(self) -> bool:
        """
        :return: true if file format supports origin based representation
        """
        return True



from abc import ABC
from io import TextIOWrapper
from typing import Any

from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.writer.base import BaseWriter


class GeoOffWriter(BaseWriter, ABC):
    """
    Writer implementation for creating Geo-Referenced OFF geometry files (.geooff)
    """

    def _write(
        self,
        file: TextIOWrapper,
        data: GeoObjectFile,
        write_binary: bool,
        random_seed: Any,
    ) -> None:
        """
        Write implementation
        :param file: target to be written
        :param data: content to be written
        :param write_binary: flag if file is a binary file
        :return:
        """
        if len(data.objects) != 1:
            raise Exception("GeoOFF can represent only one object. Minimize the data.")

        num_vertices = len(data.vertices)
        faces = []
        for obj in data.objects:
            if (
                obj.contains_scaling()
                or obj.contains_rotation()
                or obj.contains_translation()
            ):
                raise Exception(
                    "GeoOFF does not support local object transformation information."
                )
            for face in obj.faces:
                faces.append(face)
        num_faces = len(faces)
        origin_based = data.is_origin_based()
        contains_extent = data.contains_extent()
        contains_scaling = data.contains_scaling()
        contains_translation = data.contains_translation()
        contains_rotation = data.contains_rotation()
        if data.is_geo_referenced():
            header = "GeoOFF"
            if origin_based:
                header += "o"
            if contains_extent:
                header += "e"
            if contains_scaling:
                header += "s"
            if contains_translation:
                header += "t"
            if contains_rotation:
                header += "r"
            self._write_to_file(file, header, write_binary, True)
            self._write_to_file(file, data.crs, write_binary, True)
            if data.origin is not None and origin_based:
                self._write_to_file(
                    file, " ".join([str(f) for f in data.origin]), write_binary, True
                )
            if (
                data.min_extent is not None
                and data.max_extent is not None
                and contains_extent
            ):
                self._write_to_file(
                    file,
                    " ".join([str(f) for f in data.min_extent])
                    + " ".join([str(f) for f in data.max_extent]),
                    write_binary,
                    True,
                )
            if data.scaling is not None and contains_scaling:
                self._write_to_file(
                    file, " ".join([str(f) for f in data.scaling]), write_binary, True
                )
            if data.translation is not None and contains_translation:
                self._write_to_file(
                    file,
                    " ".join([str(f) for f in data.translation]),
                    write_binary,
                    True,
                )
            if data.rotation is not None and contains_rotation:
                self._write_to_file(
                    file, " ".join([str(f) for f in data.rotation]), write_binary, True
                )
        else:
            if origin_based:
                raise Exception("Origin information not supported in OFF file format")
            if contains_extent:
                raise Exception("Extent information not supported in OFF file format")
            if contains_scaling:
                raise Exception("Scaling information not supported in OFF file format")
            if contains_translation:
                raise Exception(
                    "Translation information not supported in OFF file format"
                )
            if contains_rotation:
                raise Exception("Rotation information not supported in OFF file format")
            self._write_to_file(file, "OFF", write_binary, True)

        self._write_to_file(file, f"{num_vertices} {num_faces} 0", write_binary, True)

        for v in data.vertices:
            self._write_to_file(file, " ".join([str(f) for f in v]), write_binary, True)
        for face in faces:
            s = ""
            for f in face.indices:
                s += " "
                f = int(f)
                if f > 0:
                    s += str(f - 1)
                else:
                    s += str(num_vertices - f)
            self._write_to_file(file, f"{len(face.indices)}{s}", write_binary, True)

    def get_file_type(self) -> str:
        """
        :return: the supported file type of this writer
        """
        return ".geooff"

    def supports_origin_base(self) -> bool:
        """
        :return: true if file format supports origin based representation
        """
        return True



import copy
from typing import Dict, List

from geofiles.conversion.calculation import convert_obj_index, get_center, rotate_point
from geofiles.domain.face import Face
from geofiles.domain.geo_object_file import GeoObjectFile


class Transformer:
    """
    class used to apply different transformations to given GeoObjectFile
    """

    def rotate(
        self,
        data: GeoObjectFile,
        update_extents: bool = False,
        apply_only_global: bool = True,
    ) -> GeoObjectFile:
        """
        Applies the rotation defined in the given data
        :param data: to be rotated
        :param update_extents: Flag if extents of result should be re-calculated
        :param apply_only_global: Flag that decides if only global transformations should be applied or also local ones
        :return: rotated clone of the given data
        """
        return self.transform(
            data,
            scale=False,
            translate=False,
            update_extents=update_extents,
            apply_only_global=apply_only_global,
        )

    def translate(
        self,
        data: GeoObjectFile,
        update_extents: bool = False,
        apply_only_global: bool = True,
    ) -> GeoObjectFile:
        """
        Applies the translation defined in the given data
        :param data: to be translated
        :param update_extents: Flag if extents of result should be re-calculated
        :param apply_only_global: Flag that decides if only global transformations should be applied or also local ones
        :return: rotated clone of the given data
        """
        return self.transform(
            data,
            rotate=False,
            scale=False,
            update_extents=update_extents,
            apply_only_global=apply_only_global,
        )

    def scale(
        self,
        data: GeoObjectFile,
        update_extents: bool = False,
        apply_only_global: bool = True,
    ) -> GeoObjectFile:
        """
        Applies the scaling defined in the given data
        :param data: to be rotated
        :param update_extents: Flag if extents of result should be re-calculated
        :param apply_only_global: Flag that decides if only global transformations should be applied or also local ones
        :return: scaled clone of the given data
        """
        return self.transform(
            data,
            rotate=False,
            translate=False,
            update_extents=update_extents,
            apply_only_global=apply_only_global,
        )

    @staticmethod
    def transform(
        data: GeoObjectFile,
        scale: bool = True,
        rotate: bool = True,
        translate: bool = True,
        update_extents: bool = False,
        apply_only_global: bool = True,
    ) -> GeoObjectFile:
        """
        Applies the scaling, rotation and translation for the given data
        :param data: to be transformed
        :param scale: Flag if scaling should be applied
        :param rotate: Flag if rotation should be applied
        :param translate: Flag if translation should be applied
        :param update_extents: Flag if extents of result should be re-calculated
        :param apply_only_global: Flag that decides if only global transformations should be applied or also local ones
        :return: transformed clone of the given data
        """
        if not data.is_origin_based():
            raise Exception("Function only supported for origin based representations")

        res = copy.deepcopy(data)
        scaling: List[float]
        if scale and res.scaling is not None:
            scaling = res.scaling
            res.scaling = None
        else:
            scaling = [1, 1, 1]

        rotation: List[float]
        if rotate and res.rotation is not None:
            rotation = res.rotation
            res.rotation = None
        else:
            rotation = [0, 0, 0]

        translation: List[float]
        if translate and res.translation is not None:
            translation = res.translation
            res.translation = None
        else:
            translation = [0, 0, 0]

        center = get_center(res.vertices)

        new_vertices = []

        for vertex in res.vertices:
            new_vertices.append(
                Transformer._transform_vertex(
                    vertex, center, scaling, rotation, translation
                )
            )
        res.vertices = new_vertices

        if not apply_only_global:
            center = get_center(res.vertices)
            vertex_index_mapping: Dict[str, int] = dict()
            transformed_vertices = []
            counter = 1
            for geoobj in res.objects:
                if (
                    geoobj.contains_translation()
                    or geoobj.contains_rotation()
                    or geoobj.contains_translation()
                ):
                    if geoobj.translation is not None and geoobj.contains_translation():
                        local_translation = geoobj.translation
                        geoobj.translation = None
                    else:
                        local_translation = [0, 0, 0]

                    if geoobj.rotation is not None and geoobj.contains_rotation():
                        local_rotation = geoobj.rotation
                        geoobj.rotation = None
                    else:
                        local_rotation = [0, 0, 0]

                    if geoobj.scaling is not None and geoobj.contains_scaling():
                        local_scaling = geoobj.scaling
                        geoobj.scaling = None
                    else:
                        local_scaling = [1, 1, 1]
                    new_faces = []
                    for face in geoobj.faces:
                        new_face = Face()
                        for idx in face.indices:
                            vertex = res.vertices[convert_obj_index(idx, res.vertices)]
                            transformed = Transformer._transform_vertex(
                                vertex,
                                center,
                                local_scaling,
                                local_rotation,
                                local_translation,
                            )
                            string_representation = " ".join(
                                [str(a) for a in transformed]
                            )
                            if vertex_index_mapping.get(string_representation) is None:
                                vertex_index_mapping[string_representation] = counter
                                transformed_vertices.append(transformed)
                                counter += 1
                            new_face.indices.append(
                                vertex_index_mapping[string_representation]
                            )
                        new_faces.append(new_face)
                    geoobj.faces = new_faces
                else:
                    new_faces = []
                    for face in geoobj.faces:
                        new_face = Face()
                        for idx in face.indices:
                            vertex = res.vertices[convert_obj_index(idx, res.vertices)]
                            string_representation = " ".join([str(a) for a in vertex])
                            if vertex_index_mapping.get(string_representation) is None:
                                vertex_index_mapping[string_representation] = counter
                                transformed_vertices.append(vertex)
                                counter += 1
                            new_face.indices.append(
                                vertex_index_mapping[string_representation]
                            )
                        new_faces.append(new_face)
                    geoobj.faces = new_faces
            res.vertices = transformed_vertices

        res.min_extent = []
        res.max_extent = []
        if update_extents:
            res.update_extent()

        return res

    @staticmethod
    def _transform_vertex(
        vertex: List[float],
        center: List[float],
        scaling: List[float],
        rotation: List[float],
        translation: List[float],
    ) -> List[float]:
        """
        Transforms the given vertex using the params
        :param center: Center of all vertices
        :param scaling: To be applied
        :param rotation: To be applied
        :param translation: To be applied
        :returns: transformed vertex
        """
        # recenter around (0,0,0) and scale it
        new_vertex = [
            (v_i - c_i) * s_i for v_i, c_i, s_i in zip(vertex, center, scaling)
        ]
        rotated = rotate_point(
            new_vertex, [0, 0, 0], rotation[0], rotation[1], rotation[2]
        )
        # recenter around origin
        rotated = [
            c_i + r_i + t_i for c_i, r_i, t_i in zip(center, rotated, translation)
        ]
        return rotated



import math
from abc import ABC, abstractmethod
from io import StringIO, TextIOWrapper
from typing import Any, Union

from geofiles.domain.geo_object_file import GeoObjectFile


class BaseWriter(ABC):
    """
    Base implementation for writing geo-referenced object
    """

    def write(
        self,
        file: Any,
        data: GeoObjectFile,
        write_binary: bool = False,
        append_file_type: bool = True,
        random_seed: Any = None,
    ) -> None:
        """
        Allows to create write a file at the given file position
        :param file: file to be written. Either a opened file or a path to the file
        :param data: to be written
        :param write_binary: flag if file should be written in binary style (only used if file-parameter is a path)
        :param append_file_type: flag if a writer's associated file type should be appended to the given file (only if it is a path)
        :param random_seed: may be used by the writer for e.g. IDs
        :return: None
        """
        close = False
        to_write: Any = None
        try:
            if isinstance(file, str):
                if append_file_type:
                    file += self.get_file_type()

                if write_binary:
                    to_write = open(file, "wb")
                else:
                    to_write = open(file, "w")
                close = True
            elif isinstance(file, TextIOWrapper):
                if "w" not in file.mode.lower():
                    raise Exception("Given file is not in write mode")

                if "b" in file.mode.lower():
                    write_binary = True
                to_write = file
            else:
                raise Exception(f"Can't handle {file}")

            self._write(to_write, data, write_binary, random_seed)

        finally:
            if close and to_write is not None:
                to_write.close()

    @abstractmethod
    def _write(
        self,
        file: Union[TextIOWrapper, StringIO],
        data: GeoObjectFile,
        write_binary: bool,
        random_seed: Any,
    ) -> None:
        """
        Write implementation
        :param file: target to be written
        :param data: content to be written
        :param write_binary: flag if file is a binary file
        :param random_seed: may be used by the writer for e.g. IDs
        :return:
        """
        return

    @abstractmethod
    def get_file_type(self) -> str:
        """
        :return: the supported file type of this writer
        """
        return ""

    def write_to_string(
        self, data: GeoObjectFile, write_binary: bool = False, random_seed: Any = None
    ) -> str:
        """
        Write to string implementation
        :param data: content to be written
        :param write_binary: flag if file should be written in binary style (only used if file-parameter is a path)
        :param random_seed: may be used by the writer for e.g. IDs
        :return: string representation of the given data
        """
        to_write = StringIO("")
        self._write(to_write, data, write_binary, random_seed)
        to_write.seek(0)
        return "".join(to_write.readlines())

    def supports_origin_base(self) -> bool:  # pylint: disable=R0201
        """
        :return: true if file format supports origin based representation
        """
        return False

    def _write_to_file(
        self,
        file: Union[TextIOWrapper, StringIO],
        data: Any,
        write_binary: bool,
        append_new_line: bool = False,
        encoding: str = "ascii",
    ) -> None:
        """
        Write to the given file
        :param file: to be written to
        :param data: what to be written
        :param write_binary: flag if data should be written binary ascii-encoded
        :param append_new_line: flag if new line character should be appended
        :param encoding: how data should be encoded
        :return: None
        """
        file.write(self._encode(data, write_binary, encoding))
        if append_new_line:
            file.write(self._encode("\n", write_binary, encoding))

    @staticmethod
    def _encode(data: Any, write_binary: bool, encoding: str = "ascii") -> Any:
        """
        Encode the given data
        :param data: to be encoded
        :param write_binary: flag if data should be encoded
        :param encoding: how data should be encoded
        :return: encoded data
        """
        if write_binary:
            return str(data).encode(encoding)
        return f"{data}"

    @staticmethod
    def _contains_transformation_information(data: GeoObjectFile) -> None:
        """
        Check the given data if it contains translation, rotation or scale information
        :param data: to be checked
        :return: None
        """
        if data.translation is not None:
            for t in data.translation:
                if not math.isclose(float(t), 0, rel_tol=1e-6):
                    raise Exception(
                        "Given data contains translation information. Transform data, before writing to file"
                    )

        if data.rotation is not None:
            for t in data.rotation:
                if not math.isclose(float(t), 0, rel_tol=1e-6):
                    raise Exception(
                        "Given data contains rotation information. Transform data, before writing to file"
                    )

        if data.scaling is not None:
            for t in data.scaling:
                if not math.isclose(float(t), 1, rel_tol=1e-6):
                    raise Exception(
                        "Given data contains scale information. Transform data, before writing to file"
                    )



from abc import ABC
from typing import Any, Dict, List

from geofiles.conversion.static import get_wgs_84
from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.writer.base import BaseWriter
from geofiles.writer.json_writer import JsonWriter


class GeoJsonWriter(JsonWriter, BaseWriter, ABC):
    """
    Writer implementation for creating GeoJSON geometry files
    """

    def create_json(
        self,
        data: GeoObjectFile,
        random_seed: Any = None,
    ) -> Dict[Any, Any]:
        if data.crs != get_wgs_84():
            raise Exception("GeoJSON (RFC 7946) requires WGS:84 coordinate system")

        if data.is_origin_based():
            raise Exception("Geo-referenced data must not be origin based")

        self._contains_transformation_information(data)

        res: Dict[Any, Any] = dict()
        res["type"] = "FeatureCollection"
        features: List[Any] = []
        res["features"] = features
        for obj in data.objects:
            if (
                obj.contains_scaling()
                or obj.contains_rotation()
                or obj.contains_translation()
            ):
                raise Exception(
                    "GeoJSON does not support local object transformation information"
                )
            polygon: Dict[Any, Any] = dict()
            polygon["type"] = "MultiPolygon"
            coordinates: List[Any] = []
            polygon["coordinates"] = coordinates
            for face in obj.faces:
                f = []
                for idx in face.indices:
                    vertex = data.get_vertex(idx)
                    f.append(vertex)
                f.append(data.get_vertex(face.indices[0]))
                coordinates.append([f])
            features.append(polygon)

        return res

    def get_file_type(self) -> str:
        """
        :return: the supported file type of this writer
        """
        return ".geo.json"



from typing import List


class Face:
    """
    Class that represents an individual face
    """

    def __init__(self) -> None:
        """
        Initializes a face with the following attributes:
        - indices: The indices of all vertices that represent the polygonal boundary of this face
        - normal_indices: The indices of the normals of this face
        - texture_coordinates: Indices of the the texture_coordinates of this face

        Note that indices use the obj ordering with start index 1 and indexes < 0 for tail access (so -1 == vertices[len(vertices)-1]
        """
        self.indices: List[int] = []
        self.normal_indices: List[int] = []
        self.texture_coordinates: List[int] = []



from abc import ABC
from io import TextIOWrapper
from typing import Any, List

from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.writer.base import BaseWriter


class GeoStlWriter(BaseWriter, ABC):
    """
    Writer implementation for creating Geo-Referenced STL geometry files (.geostl)
    """

    def __init__(self) -> None:
        self.stl_name = ""

    def _write(
        self,
        file: TextIOWrapper,
        data: GeoObjectFile,
        write_binary: bool,
        random_seed: Any,
    ) -> None:
        """
        Write implementation
        :param file: target to be written
        :param data: content to be written
        :param write_binary: flag if file is a binary file
        :return:
        """
        if len(data.objects) != 1:
            raise Exception("GeoSTL can represent only one object. Minimize the data.")

        self._contains_transformation_information(data)

        if data.is_geo_referenced():
            origin = ""
            if data.is_origin_based() and data.origin is not None:
                for c in data.origin:
                    origin += str(c) + " "
            self._write_to_file(
                file,
                f"geosolid {data.crs} {origin} {self.stl_name}",
                write_binary,
                True,
            )
        else:
            self._write_to_file(file, f"solid {self.stl_name}", write_binary, True)

        for obj in data.objects:
            if (
                obj.contains_scaling()
                or obj.contains_rotation()
                or obj.contains_translation()
            ):
                raise Exception(
                    "GeoSTL does not support local object transformation information"
                )
            for face in obj.faces:
                if len(face.normal_indices) != 0:
                    normal: List[Any] = [0, 0, 0]
                    for idx in face.normal_indices:
                        n = data.get_normal(idx)
                        normal = [float(x) + float(y) for x, y in zip(normal, n)]
                    normal = [str(x / len(face.normal_indices)) for x in normal]

                    self._write_to_file(
                        file, f" facet normal {' '.join(normal)}", write_binary, True
                    )
                else:
                    self._write_to_file(file, " facet", write_binary, True)
                self._write_to_file(file, "  outer loop", write_binary, True)
                for idx in face.indices:
                    vertex = [str(a) for a in data.get_vertex(idx)]
                    self._write_to_file(
                        file, f"   vertex {' '.join(vertex)}", write_binary, True
                    )
                self._write_to_file(file, "  endloop", write_binary, True)
                self._write_to_file(file, " endfacet", write_binary, True)
        if data.is_geo_referenced():
            self._write_to_file(file, "endgeosolid", write_binary, True)
        else:
            self._write_to_file(file, "endsolid", write_binary, True)

    def get_file_type(self) -> str:
        """
        :return: the supported file type of this writer
        """
        return ".geostl"

    def supports_origin_base(self) -> bool:
        """
        :return: true if file format supports origin based representation
        """
        return True



import xml.etree.ElementTree as ET
from abc import ABC
from typing import Any, Dict, List

from geofiles.conversion.static import get_wgs_84
from geofiles.domain.face import Face
from geofiles.domain.geo_object import GeoObject
from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.reader.base import BaseReader
from geofiles.reader.xml_reader import XmlReader


class KmlReader(XmlReader, BaseReader, ABC):
    """
    Reader implementation for KML files
    Note: That only Solids containing Polygons are supported. Additionally, only the Exterior Linearrings are considered.
    """

    def __init__(self):
        """
        unique_vertices: defines that read vertices have to be unique
        """
        self.unique_vertices = False

    def read_xml(self, xml: ET.Element) -> GeoObjectFile:
        result = GeoObjectFile()
        result.crs = get_wgs_84()

        self.remove_namespaces(xml)
        vertex_list: List[List[Any]] = []
        vertex_indices: Dict[str, int] = dict()
        placemarks = xml.findall(".//Placemark")
        for placemark in placemarks:
            geo_object = GeoObject()
            result.objects.append(geo_object)
            names = placemark.findall(".//name")
            if len(names) == 1 and names[0].text:
                geo_object.name = names[0].text
            polygons = placemark.findall(".//Polygon")
            for polygon in polygons:
                coordinates = polygon.findall(
                    "./outerBoundaryIs/LinearRing/coordinates"
                )
                if len(coordinates) == 1 and coordinates[0].text:
                    face_object = Face()
                    splitted_coordinates = coordinates[0].text.split("\n")
                    splitted_coordinates.pop()
                    for splitted_coordinate in [a for a in splitted_coordinates if a]:
                        splitted = [
                            float(a) for a in splitted_coordinate.split(" ") if a
                        ]
                        self._filter_faces(
                            self.unique_vertices,
                            splitted,
                            face_object,
                            vertex_list,
                            vertex_indices,
                        )
                    geo_object.faces.append(face_object)

        result.vertices = vertex_list
        return result



import copy
from typing import List, Optional

from geofiles.conversion.static import is_not_none_nor_empty, update_min_max
from geofiles.domain.geo_object import GeoObject


class GeoObjectFile:
    """
    Basic class representing a file containing geo referenced object
    """

    def __init__(self) -> None:
        """
        Initializes a GeoObjectFile with the following attributes:
        - crs: name (string) of the used coordinate reference system
        - origin: geo-referenced origin (tuple) of the geo objects
        - translation: tuple containing the global translation of origin-based geo objects
        - rotation: tuple containing the global rotation of origin-based geo objects
        - scaling: tuple containing the global scaling of origin-based geo objects
        - objects: List of all geo-objects within this GeoObjectFile
        - vertices: List of all vertices within this GeoObjectFile
        - normals: List of all normals within this GeoObjectFile
        - texture_coordinates: List of all texture_coordinates within this GeoObjectFile
        - min_extent: minimal geographical extent of the vertices
        - max_extent: maximal geographical extent of the vertices
        """
        self.crs: Optional[str] = None
        self.origin: Optional[List[float]] = None
        self.translation: Optional[List[float]] = None
        self.rotation: Optional[List[float]] = None
        self.scaling: Optional[List[float]] = None
        self.objects: List[GeoObject] = []
        self.vertices: List[List[float]] = []
        self.normals: List[List[float]] = []
        self.texture_coordinates: List[List[float]] = []
        self.min_extent: Optional[List[float]] = []
        self.max_extent: Optional[List[float]] = []

    def is_origin_based(self) -> bool:
        """
        Checks if this GeoObjectFile is based on a geo referenced origin, or if all vertices are geo referenced
        :return: true iff file uses a origin
        """
        return self.origin is not None

    def is_geo_referenced(self) -> bool:
        """
        :return: true iff file is geo-referenced or a local file
        """
        return self.crs is not None

    def get_vertex(self, idx: int) -> List[float]:
        """
        Returns the vertex at the given index (Note: indices are .obj style starting with 1 and tail indices < 0
        :param idx: idx to access
        :return: vertex at given index
        """
        return self._access_idx(self.vertices, idx)

    def get_normal(self, idx: int) -> List[float]:
        """
        Returns the normal vertex at the given index (Note: indices are .obj style starting with 1 and tail indices < 0
        :param idx: idx to access
        :return: normal vertex at given index
        """
        return self._access_idx(self.normals, idx)

    @staticmethod
    def _access_idx(list_to_access: List[List[float]], idx: int) -> List[float]:
        """
        Access the given list using the given index (Note: indices are .obj style starting with 1 and tail indices < 0)
        :param list_to_access: list to be accessed
        :param idx: index
        :return: element in list at given index
        """
        idx = int(idx)

        if idx > 0:
            return list_to_access[idx - 1]

        if idx < 0:
            return list_to_access[idx]

        raise Exception(f"Non valid index {idx}")

    def contains_extent(self) -> bool:
        """
        Checks if this geo-referenced file contains extent information
        """
        return is_not_none_nor_empty(self.min_extent) and is_not_none_nor_empty(
            self.max_extent
        )

    def contains_scaling(self) -> bool:
        """
        Checks if this geo-referenced file contains global scaling information
        """
        return is_not_none_nor_empty(self.scaling)

    def contains_translation(self) -> bool:
        """
        Checks if this geo-referenced file contains global translation information
        """
        return is_not_none_nor_empty(self.translation)

    def contains_rotation(self) -> bool:
        """
        Checks if this geo-referenced file contains global rotation information
        """
        return is_not_none_nor_empty(self.rotation)

    def update_extent(self) -> None:
        """
        Updates the min and max extent values of this geo-referenced object file, it does not consider if the file is origin based,
        nor any transformation information. For a more advanced functionality use the ExtentCalculator class
        """
        if len(self.vertices) > 0:
            min_extent = list(copy.deepcopy(self.vertices[0]))
            max_extent = list(copy.deepcopy(self.vertices[0]))

            for vertex in self.vertices:
                update_min_max(vertex, min_extent, max_extent)

            self.min_extent = min_extent
            self.max_extent = max_extent

    def minimize(self, name: Optional[str] = None) -> None:
        """
        Minimizes this GeoObjectFile to one single object as required for GeoOFF and GeoPLY. Also eliminates duplicated faces
        :param name: Name for the single object, if None the name of the first object is used
        """
        geoobject = GeoObject()

        use_first_elements_name = False
        if name is not None:
            geoobject.name = name
        else:
            use_first_elements_name = True

        face_set = set()
        for old_object in self.objects:
            if (
                old_object.contains_translation()
                or old_object.contains_scaling()
                or old_object.contains_rotation()
            ):
                raise Exception(
                    "Can not minimize GeoObjectFile containing objects with local transformation."
                )
            if use_first_elements_name:
                geoobject.name = old_object.name
                use_first_elements_name = False

            for face in old_object.faces:
                face_set.add(face)

        geoobject.faces = list(face_set)

        self.objects.clear()
        self.objects.append(geoobject)



import copy
from typing import Any, List

from geofiles.conversion.origin_converter import OriginConverter
from geofiles.domain.geo_object_file import GeoObjectFile


class LocalConverter:
    """
    Converter used to convert a geo referenced file to a local file or vice versa
    """

    @staticmethod
    def from_local(
        data: GeoObjectFile,
        crs: str,
        origin: List[Any],
        origin_based: bool = True,
        update_extent: bool = False,
    ) -> GeoObjectFile:
        """
        Converts the given object file with a local coordinate system to a geo-referenced representation
        :param data: to be converted
        :param crs: target crs
        :param origin: origin point used for conversion
        :param origin_based: flag that signals if the final file should be origin based or should use geo-referenced vertices
        :param update_extent: If true, extent information of the converted data is determined
        :return: geo-referenced file based on the given origin
        """
        if data.crs is not None:
            raise Exception("Given data is already geo-referenced")
        res = copy.deepcopy(data)
        res.crs = crs
        res.origin = origin
        if not origin_based:
            converter = OriginConverter()
            res = converter.from_origin(res, update_extent=update_extent)
        elif update_extent:
            res.update_extent()

        return res

    @staticmethod
    def to_local(data: GeoObjectFile, update_extent: bool = False) -> GeoObjectFile:
        """
        Converts the given geo-referenced file to a local representation
        :param data: to be converted
        :param update_extent: If true, extent information of the converted data is determined
        :return:
        """
        if not data.is_geo_referenced():
            raise Exception("Given data is already in local coordinates")

        res: GeoObjectFile
        if not data.is_origin_based():
            converter = OriginConverter()
            res = converter.to_origin(data, update_extent=update_extent)
        else:
            res = copy.deepcopy(data)
            if update_extent:
                res.update_extent()

        res.origin = None
        res.crs = None

        return res



import xml.etree.ElementTree as ET
from abc import ABC
from typing import Any, Dict, List

from geofiles.domain.face import Face
from geofiles.domain.geo_object import GeoObject
from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.reader.base import BaseReader
from geofiles.reader.xml_reader import XmlReader


class GmlReader(XmlReader, BaseReader, ABC):
    """
    Reader implementation for GML files
    Note: That only Solids containing Polygons are supported. Additionally, only the Exterior Linearrings are considered.
    """

    def __init__(self):
        """
        unique_vertices: defines that read vertices have to be unique
        """
        self.unique_vertices = False

    def read_xml(self, xml: ET.Element) -> GeoObjectFile:
        result = GeoObjectFile()
        self.remove_namespaces(xml)

        solids = xml.findall(".//Solid")

        vertex_list: List[List[Any]] = []
        vertex_indices: Dict[str, int] = dict()

        if len(solids) > 0:
            first_solid = solids[0]
            result.crs = self._get_attribute(first_solid, "srsName")
            for solid in solids:
                if self._get_attribute(solid, "srsName") != result.crs:
                    raise Exception(
                        "Found non uniform CRS definition in Solid elements. Currently not supported in this implementation."
                    )
                geo_object = GeoObject()

                polygons = solid.findall(".//Polygon")

                for polygon in polygons:
                    poslists = polygon.findall(".//exterior/LinearRing/posList")
                    face_object = Face()
                    for poslist in poslists:
                        if poslist is not None and poslist.text is not None:
                            splits = poslist.text.split(" ")
                            splits.pop()
                            for split in splits:
                                coordinate = [float(a) for a in split.split(",")]
                                self._filter_faces(
                                    self.unique_vertices,
                                    coordinate,
                                    face_object,
                                    vertex_list,
                                    vertex_indices,
                                )
                    geo_object.faces.append(face_object)
                result.objects.append(geo_object)

        result.vertices = vertex_list
        return result



import uuid
from abc import ABC
from typing import Any, Dict, List

from geofiles.domain.file_version import CityJsonVersion
from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.writer.base import BaseWriter
from geofiles.writer.json_writer import JsonWriter


class CityJsonWriter(JsonWriter, BaseWriter, ABC):
    """
    Writer implementation for creating GeoJSON geometry files
    """

    def __init__(self, version: CityJsonVersion = CityJsonVersion.V1_1):
        """
        :param version: The CityJSON version to be used
        """
        self.version = version

    def create_json(
        self,
        data: GeoObjectFile,
        random_seed: Any = None,
    ) -> Dict[Any, Any]:
        if data.is_origin_based():
            raise Exception("Geo-referenced data must not be origin based")

        if data.contains_rotation():
            raise Exception("CityJSON does not support rotational information")

        num_vertices = len(data.vertices)

        res: Dict[Any, Any] = dict()
        res["type"] = "CityJSON"

        if self.version == CityJsonVersion.V1_1:
            res["version"] = "1.1"
        else:
            res["version"] = "1.0"

        metadata: Dict[Any, Any] = dict()
        add_metadata = False
        if data.crs is not None:
            add_metadata = True
            metadata["referenceSystem"] = data.crs

        transform = dict()
        set_scale = False
        if data.contains_scaling():
            scale = data.scaling
            set_scale = True
        elif self.version != CityJsonVersion.V1_0:
            # scale is mandatory in CityJSON 1.1
            scale = [1, 1, 1]
            set_scale = True

        set_translate = False
        if data.contains_translation():
            translate = data.translation
            set_translate = True
        elif self.version != CityJsonVersion.V1_0:
            # translation is mandatory in CityJSON 1.1
            translate = [0, 0, 0]
            set_translate = True

        if set_scale:
            transform["scale"] = scale
        if set_translate:
            transform["translate"] = translate

        if len(transform) != 0:
            res["transform"] = transform

        if (
            data.min_extent is not None
            and data.max_extent is not None
            and len(data.min_extent) > 0
            and len(data.max_extent) > 0
        ):
            add_metadata = True
            metadata["geographicalExtent"] = data.min_extent + data.max_extent

        if add_metadata:
            res["metadata"] = metadata
        objects: Dict[Any, Any] = dict()
        res["CityObjects"] = objects
        for obj in data.objects:
            if (
                obj.contains_scaling()
                or obj.contains_rotation()
                or obj.contains_translation()
            ):
                raise Exception(
                    "CityJSON does not support local object transformation information"
                )
            cityobject: Dict[Any, Any] = dict()
            if obj.name == "" or obj.name is None:
                if random_seed is None:
                    objects[str(uuid.uuid4())] = cityobject
                else:
                    objects[str(uuid.UUID(int=random_seed, version=4))] = cityobject
            else:
                objects[obj.name] = cityobject
            cityobject["type"] = "GenericCityObject"
            geometries: List[Any] = []
            cityobject["geometry"] = geometries
            geometry: Dict[Any, Any] = dict()
            geometries.append(geometry)
            geometry["type"] = "MultiSurface"
            geometry["lod"] = 1
            boundaries: List[Any] = []
            geometry["boundaries"] = boundaries
            for face in obj.faces:
                boundary = []
                for idx in face.indices:
                    idx = int(idx)
                    if idx < 0:
                        boundary.append(num_vertices - idx)
                    else:
                        boundary.append(idx - 1)
                boundaries.append([boundary])
        res["vertices"] = data.vertices

        return res

    def get_file_type(self) -> str:
        """
        :return: the supported file type of this writer
        """
        return ".city.json"



from typing import Any, List, Optional


def get_wgs_84() -> str:
    """
    :return: name of the WGS 84 coordinate system
    """
    return "urn:ogc:def:crs:OGC:2:84"


def get_epsg_4326() -> str:
    """
    :return: name of the EPSG 4326 coordinate system
    """
    return "urn:ogc:def:crs:EPSG::4326"


def get_lon_lat(vertex: List[float], coordinate_system: str) -> List[float]:
    """
    Method for accessing longitude and latitude
    :param vertex: WGS84 or EPSG4236 vertex
    :param coordinate_system: coordinate system of the vertex
    :return: longitude, latitude
    """
    if coordinate_system == get_wgs_84():
        return [vertex[0], vertex[1]]

    if coordinate_system == get_epsg_4326():
        return [vertex[1], vertex[0]]

    raise Exception(f"Not supported crs {coordinate_system}")


def update_min_max(
    vertex: List[float],
    min_extent: Optional[List[float]],
    max_extent: Optional[List[float]],
) -> None:
    """
    Update the min and max extent information based on the vertex
    :param vertex: input
    :param min_extent: current min extent
    :param max_extent: current max extent
    :return: None
    """
    if min_extent is None or max_extent is None:
        return

    for idx, elem in enumerate(vertex):
        min_value = min_extent[idx]
        max_value = max_extent[idx]
        if min_value is None or elem < min_value:
            min_extent[idx] = elem
        if max_value is None or elem > max_value:
            max_extent[idx] = elem


def is_not_none_nor_empty(to_check: Optional[List[Any]]) -> bool:
    """
    Checks if the given optional list is neither None nor empty
    :param to_check: List to be checked
    :results: true if list is not none and list is not empty
    """
    return to_check is not None and len(to_check) > 0



import xml.etree.ElementTree as ET
from abc import ABC
from typing import Any

from geofiles.conversion.static import get_wgs_84
from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.writer.base import BaseWriter
from geofiles.writer.xml_writer import XmlWriter


class KmlWriter(XmlWriter, BaseWriter, ABC):
    """
    Writer implementation for creating KML geometry files
    """

    def create_xml(
        self, data: GeoObjectFile, random_seed: Any = None
    ) -> ET.ElementTree:
        self._contains_transformation_information(data)

        if data.crs != get_wgs_84():
            raise Exception("Kml requires WGS:84 coordinate system")

        if data.is_origin_based():
            raise Exception("Geo-referenced data must not be origin based")

        root = ET.Element("kml")
        root.attrib["xmlns"] = "http://www.opengis.net/kml/2.2"
        for obj in data.objects:
            if (
                obj.contains_scaling()
                or obj.contains_rotation()
                or obj.contains_translation()
            ):
                raise Exception(
                    "KML does not support local object transformation information"
                )
            placemark = ET.Element("Placemark")
            root.append(placemark)
            name = ET.Element("name")
            name.text = obj.name
            placemark.append(name)
            for face in obj.faces:
                polygon = ET.Element("Polygon")
                placemark.append(polygon)
                extrude = ET.Element("extrude")
                extrude.text = "0"
                polygon.append(extrude)
                altitude_mode = ET.Element("altitudeMode")
                altitude_mode.text = "absolute"
                polygon.append(altitude_mode)
                outer_boundary = ET.Element("outerBoundaryIs")
                polygon.append(outer_boundary)
                linear_ring = ET.Element("LinearRing")
                outer_boundary.append(linear_ring)
                coordinates = ET.Element("coordinates")
                coordinates.text = "\n"
                linear_ring.append(coordinates)
                for idx in face.indices:
                    coords = data.get_vertex(idx)
                    for coord in coords:
                        coordinates.text += str(coord) + " "
                    coordinates.text += "\n"

                coords = data.get_vertex(face.indices[0])
                for coord in coords:
                    coordinates.text += str(coord) + " "

        return ET.ElementTree(root)

    def get_file_type(self) -> str:
        """
        :return: the supported file type of this writer
        """
        return ".kml"



from abc import ABC
from typing import Any, Dict, Iterable

from geofiles.domain.face import Face
from geofiles.domain.geo_object import GeoObject
from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.reader.base import BaseReader


class GeoStlReader(BaseReader, ABC):
    """
    Reader implementaiton for geo-referenced .stl files (.geostl)
    """

    def _read(self, file: Iterable[str]) -> GeoObjectFile:

        res = GeoObjectFile()
        obj = GeoObject()
        res.objects.append(obj)
        vertices: Dict[Any, Any] = dict()
        cnt = 1
        current_face: Face
        for line in file:
            trimmed = line.strip()
            if not trimmed:
                continue
            trimmed = " ".join(trimmed.split())
            is_geo_solid = trimmed.startswith("geosolid")
            if is_geo_solid or trimmed.startswith("solid"):
                splits = trimmed.split(" ")
                if is_geo_solid:
                    res.crs = splits[1]
                list_len = len(splits)
                if list_len in (5, 6):
                    res.origin = [float(a) for a in splits[2:5]]
                    if list_len == 6:
                        obj.name = splits[-1]

                if list_len == 2:
                    obj.name = splits[-1]
            elif trimmed.startswith("facet"):
                current_face = Face()
                obj.faces.append(current_face)
            elif trimmed.startswith("vertex"):
                v = trimmed[7:]
                idx = vertices.get(v)
                if idx is None:
                    vertices[v] = cnt
                    idx = cnt
                    cnt += 1
                current_face.indices.append(idx)

        for vs in vertices:
            splits = vs.split(" ")
            x = float(splits[0])
            y = float(splits[1])
            z = float(splits[2])
            res.vertices.append([x, y, z])

        return res



import xml.etree.ElementTree as ET
from abc import ABC, abstractmethod
from io import TextIOWrapper
from typing import Any

from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.writer.base import BaseWriter


class XmlWriter(BaseWriter, ABC):
    """
    Basic writer implementations for xml based file formats
    """

    @abstractmethod
    def create_xml(
        self, data: GeoObjectFile, random_seed: Any = None
    ) -> ET.ElementTree:
        """
        Creates a xml representation of the given GeoObjectFile
        :param data: to be written to xml
        :param random_seed: used for id generation
        """
        return ET.ElementTree()

    def write_to_string(
        self, data: GeoObjectFile, write_binary: bool = False, random_seed: Any = None
    ) -> str:
        et = self.create_xml(data, random_seed)
        encoding = "ascii" if write_binary else "utf8"
        return str(
            ET.tostring(et.getroot(), encoding=encoding, method="xml").decode(encoding)
        )

    def _write(
        self,
        file: TextIOWrapper,
        data: GeoObjectFile,
        write_binary: bool,
        random_seed: Any,
    ) -> None:
        if "b" not in file.mode.lower():
            raise Exception("File must be opened in binary mode")
        et = self.create_xml(data, random_seed)
        et.write(file, encoding="ascii", xml_declaration=True)



import uuid
import xml.etree.ElementTree as ET
from abc import ABC
from typing import Any

from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.writer.base import BaseWriter
from geofiles.writer.xml_writer import XmlWriter


class GmlWriter(XmlWriter, BaseWriter, ABC):
    """
    Writer implementation for creating GML geometry files
    """

    def create_xml(
        self, data: GeoObjectFile, random_seed: Any = None
    ) -> ET.ElementTree:
        if data.is_origin_based():
            raise Exception("Geo-referenced data must not be origin based")

        self._contains_transformation_information(data)

        if data.crs is None:
            raise Exception("File must be geo-referenced")

        if data.is_origin_based():
            raise Exception("Geo-referenced data must not be origin based")

        attributes = dict()
        attributes["xmlns:gml"] = "http://www.opengis.net/gml/3.2"
        root = ET.Element("root", attributes)
        for obj in data.objects:
            if (
                obj.contains_scaling()
                or obj.contains_rotation()
                or obj.contains_translation()
            ):
                raise Exception(
                    "GML does not support local object transformation information"
                )
            solid = ET.Element("gml:Solid")
            solid.attrib = dict()
            solid.attrib["srsName"] = data.crs
            solid.attrib["srsDimension"] = "3"
            root.append(solid)

            exterior = ET.Element("gml:exterior")
            solid.append(exterior)
            composite_surface = ET.Element("gml:CompositeSurface")
            exterior.append(composite_surface)
            surface_member = ET.Element("gml:surfaceMember")
            composite_surface.append(surface_member)
            for face in obj.faces:
                polygon = ET.Element("gml:Polygon")
                polygon.attrib = dict()
                if random_seed is None:
                    polygon_id = str(uuid.uuid4())
                else:
                    polygon_id = str(uuid.UUID(int=random_seed, version=4))
                polygon.attrib["gml:id"] = polygon_id
                surface_member.append(polygon)
                boundaries = ET.Element("gml:exterior")
                polygon.append(boundaries)
                linearring = ET.Element("gml:LinearRing")
                boundaries.append(linearring)
                coordinates = ET.Element("gml:posList")
                linearring.append(coordinates)
                coordinates.text = ""
                for idx in face.indices:
                    vertex = [str(a) for a in data.get_vertex(idx)]
                    coordinates.text += ",".join(vertex)
                    coordinates.text += " "

                vertex = [str(a) for a in data.get_vertex(face.indices[0])]
                coordinates.text += ",".join(vertex)

        return ET.ElementTree(root)

    def get_file_type(self) -> str:
        """
        :return: the supported file type of this writer
        """
        return ".gml"



import copy
import math
from math import cos, pi, radians, sin
from typing import Any, List


def get_distant_point(x0: float, y0: float, d: float, theta: float) -> List[float]:
    """
    Get a new point based on the start position, distance and angle
    :param x0: start x
    :param y0: start y
    :param d: distance to new point
    :param theta: angle to new point
    :return:
    """
    theta_rad = pi / 2 - radians(theta)
    return [x0 + d * cos(theta_rad), y0 + d * sin(theta_rad)]


def get_point_distance(point1: List[Any], point2: List[Any]) -> float:
    """
    Calculates the distance between the two given points
    :param point1: first point
    :param point2: second point
    :return: distance between both points
    """
    return math.sqrt(
        sum((float(px) - float(qx)) ** 2.0 for px, qx in zip(point1, point2))
    )


def get_angle_between_points(point1: List[Any], point2: List[Any]) -> float:
    """
    Calculates the angle between two points
    :param point1: first point
    :param point2: second point
    :return: angle between both points
    """
    r = (
        180.0
        / math.pi
        * math.atan2(
            float(point2[0]) - float(point1[0]), float(point2[1]) - float(point1[1])
        )
    )

    if r < 0:
        r %= 360

    return r


def get_center(vertices: List[Any]) -> List[Any]:
    """
    Calculates the center point for the given vertices
    :param vertices: for which center should be calculated
    :return: center
    """
    if len(vertices) == 0:
        raise Exception("Given vertices are empty")

    center: List[float] = []
    cnt = 0
    for vertex in vertices:
        if len(center) == 0:
            center = copy.deepcopy(vertex)
        else:
            for idx, v in enumerate(vertex):
                center[idx] += float(v)
        cnt += 1

    for idx in range(0, len(center)):  # pylint: disable=C0200
        center[idx] /= cnt

    return center


def rotate_point(
    point: List[float],
    origin: List[float],
    roll: float,
    pitch: float,
    yaw: float,
) -> List[float]:
    """
    Rotates the given 3D point around the given origin
    :param point: to be rotated
    :param origin: rotation center
    :param roll: x-rotation
    :param pitch: y-rotation
    :param yaw: z-rotation
    :return: new point representing the rotation result
    """
    curr_x = point[0] - origin[0]
    curr_y = point[1] - origin[1]
    curr_z = point[2] - origin[2]

    r_roll = math.radians(roll)
    r_pitch = math.radians(pitch)
    r_yaw = math.radians(yaw)

    cosa = math.cos(r_yaw)
    sina = math.sin(r_yaw)
    cosb = math.cos(r_pitch)
    sinb = math.sin(r_pitch)
    cosc = math.cos(r_roll)
    sinc = math.sin(r_roll)

    axx = cosa * cosb
    axy = cosa * sinb * sinc - sina * cosc
    axz = cosa * sinb * cosc + sina * sinc
    ayx = sina * cosb
    ayy = sina * sinb * sinc + cosa * cosc
    ayz = sina * sinb * cosc - cosa * sinc
    azx = -sinb
    azy = cosb * sinc
    azz = cosb * cosc

    xr = (axx * curr_x + axy * curr_y + axz * curr_z) + origin[0]
    yr = (ayx * curr_x + ayy * curr_y + ayz * curr_z) + origin[1]
    zr = (azx * curr_x + azy * curr_y + azz * curr_z) + origin[2]

    return [xr, yr, zr]


def convert_obj_index(idx: int, list_input: Any) -> int:
    """
    Method for converting an obj based index to a python like index
    :param idx: to be converted
    :param list_input: length of referenced list or the list itself
    :return: converted idx
    """
    if idx > 0:
        return idx - 1

    list_len: int
    if isinstance(list_input, int):
        list_len = list_input
    elif isinstance(list_input, list):
        list_len = len(list_input)
    else:
        raise Exception(f"Not supported type {type(list_input)} of parameter l")

    if idx < 0:
        return list_len + idx

    raise Exception("Index 0 not supported in OBJ")



from abc import ABC
from io import TextIOWrapper
from typing import Any, List

from geofiles.domain.geo_object import GeoObject
from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.writer.base import BaseWriter


class GeoObjWriter(BaseWriter, ABC):
    """
    Class for writing Geo-Referenced .obj files (.geoobj)
    """

    def _write(
        self,
        file: TextIOWrapper,
        data: GeoObjectFile,
        write_binary: bool,
        random_seed: Any,
    ) -> None:
        """
        Write implementation
        :param file: target to be written
        :param data: content to be written
        :param write_binary: flag if file is a binary file
        :return:
        """
        geo_referenced = data.is_geo_referenced()
        if data.crs is not None:
            self._write_to_file(file, "crs ", write_binary)
            self._write_to_file(file, data.crs, write_binary, True)

        if geo_referenced and data.origin is not None:
            self._write_to_file(
                file,
                "or " + " ".join([str(a) for a in data.origin]),
                write_binary,
                True,
            )

        if geo_referenced and data.scaling is not None:
            self._write_to_file(
                file,
                "sc " + " ".join([str(a) for a in data.scaling]),
                write_binary,
                True,
            )

        if geo_referenced and data.translation is not None:
            self._write_to_file(
                file,
                "t " + " ".join([str(a) for a in data.translation]),
                write_binary,
                True,
            )

        if geo_referenced and data.rotation is not None:
            self._write_to_file(
                file,
                "r " + " ".join([str(a) for a in data.rotation]),
                write_binary,
                True,
            )

        if (
            data.contains_extent()
            and data.min_extent is not None
            and data.max_extent is not None
        ):
            self._write_to_file(
                file,
                "e "
                + " ".join([str(a) for a in data.min_extent])
                + " "
                + " ".join([str(a) for a in data.max_extent]),
                write_binary,
                True,
            )

        self._write_coordinates(data.vertices, file, "v ", write_binary)
        self._write_coordinates(data.normals, file, "vn ", write_binary)
        self._write_coordinates(data.texture_coordinates, file, "vt ", write_binary)

        for geoobject in data.objects:
            if len(geoobject.faces) == 0:
                self._write_to_file(file, f"g {geoobject.name}", write_binary, True)
                if geo_referenced:
                    self._write_transformation(geoobject, file, write_binary)
            else:
                self._write_to_file(file, f"o {geoobject.name}", write_binary, True)
                if geo_referenced:
                    self._write_transformation(geoobject, file, write_binary)
                for f in geoobject.faces:
                    self._write_to_file(file, "f ", write_binary)
                    contains_textures = len(f.texture_coordinates) != 0
                    contains_normals = len(f.normal_indices) != 0
                    index_len = len(f.indices)
                    for i, idx in enumerate(f.indices):
                        self._write_to_file(file, idx, write_binary)

                        if contains_textures or contains_normals:
                            self._write_to_file(file, "/", write_binary)

                        if contains_textures:
                            self._write_to_file(
                                file, f.texture_coordinates[i], write_binary
                            )

                        if contains_normals:
                            self._write_to_file(file, "/", write_binary)
                            self._write_to_file(file, f.normal_indices[i], write_binary)

                        if i < index_len - 1:
                            self._write_to_file(file, " ", write_binary)

                    self._write_to_file(file, "", write_binary, True)

    def _write_transformation(
        self, geoobj: GeoObject, file: TextIOWrapper, write_binary: bool
    ) -> None:
        """
        Write the transformation information of the geoobject to the file
        :param geoobj: for which transformation should be written
        :param file: to which the coordinate will be written
        :param write_binary: check if ascii mode is used
        :return:
        """
        if geoobj.scaling is not None and geoobj.contains_scaling():
            self._write_to_file(
                file,
                "sc " + " ".join([str(a) for a in geoobj.scaling]),
                write_binary,
                True,
            )

        if geoobj.translation is not None and geoobj.contains_translation():
            self._write_to_file(
                file,
                "t " + " ".join([str(a) for a in geoobj.translation]),
                write_binary,
                True,
            )

        if geoobj.rotation is not None and geoobj.contains_rotation():
            self._write_to_file(
                file,
                "r " + " ".join([str(a) for a in geoobj.rotation]),
                write_binary,
                True,
            )

    def _write_coordinates(
        self,
        coordinates: List[List[Any]],
        file: TextIOWrapper,
        prefix: str,
        write_binary: bool,
    ) -> None:
        """
        Write the given coordinates to the file
        :param coordinates: to be written
        :param file: to which the coordinate will be written
        :param prefix: used to identify the coordinate type
        :param write_binary: check if ascii mode is used
        :return:
        """
        for v in coordinates:
            self._write_to_file(file, prefix, write_binary)
            self._write_to_file(file, " ".join([str(a) for a in v]), write_binary)
            self._write_to_file(file, "", write_binary, True)

    def get_file_type(self) -> str:
        """
        :return: the supported file type of this writer
        """
        return ".geoobj"

    def supports_origin_base(self) -> bool:
        """
        :return: true if file format supports origin based representation
        """
        return True



import xml.etree.ElementTree as ET
from abc import ABC, abstractmethod
from typing import Iterable, Optional

from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.reader.base import BaseReader


class XmlReader(BaseReader, ABC):
    """
    Base class for reading xml-based files
    """

    def _read(self, file: Iterable[str]) -> GeoObjectFile:
        xml_file = "".join(file)
        tree = ET.fromstring(xml_file)

        return self.read_xml(tree)

    @abstractmethod
    def read_xml(self, xml: ET.Element) -> GeoObjectFile:
        """
        Read implementation for xml based files
        :param xml: parsed element tree representation
        :return: Domain representation of the GeoObjectFile
        """
        return GeoObjectFile()

    def remove_namespaces(self, el: ET.Element) -> None:
        """
        Recursively search this element tree, removing namespaces.
        Source: https://stackoverflow.com/a/32552776
        :param el: Element for which namespace should be removed
        :return: None
        """
        if el.tag.startswith("{"):
            el.tag = el.tag.split("}", 1)[1]  # strip namespace
        for k in el.attrib.keys():
            if k.startswith("{"):
                k2 = k.split("}", 1)[1]
                el.attrib[k2] = el.attrib[k]
                del el.attrib[k]
        for child in el:
            self.remove_namespaces(child)

    @staticmethod
    def _get_attribute(xml: ET.Element, attribute_name: str) -> Optional[str]:
        """
        Get an attribute by name ignoring the namespace
        :param xml: element containing the attribute
        :param attribute_name: Name of the searched attribute
        :return: The value of the attribute if available, else None
        """
        for key, value in xml.attrib.items():
            if key.endswith(attribute_name):
                return value

        return None



# pylint: skip-file

"""
Size comparison using multiple obj base models.
You can find examples e.g. here: https://github.com/alecjacobson/common-3d-test-models
"""
import logging
import os
from os import listdir
from os.path import isfile, join
from typing import Dict

from geofiles.conversion.local_converter import LocalConverter
from geofiles.reader.geo_obj_reader import GeoObjReader
from geofiles.writer.base import BaseWriter
from geofiles.writer.city_json_writer import CityJsonWriter
from geofiles.writer.geo_json_writer import GeoJsonWriter
from geofiles.writer.geo_obj_writer import GeoObjWriter
from geofiles.writer.geo_off_writer import GeoOffWriter
from geofiles.writer.geo_ply_writer import GeoPlyWriter
from geofiles.writer.geo_stl_writer import GeoStlWriter
from geofiles.writer.geo_vrml_writer import GeoVrmlWriter
from geofiles.writer.gml_writer import GmlWriter
from geofiles.writer.kml_writer import KmlWriter

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)
    path = input("Path to sample .obj files: ")
    target_folder = join(path, "target")
    os.makedirs(target_folder, exist_ok=True)
    inputfiles = [
        f for f in listdir(path) if isfile(join(path, f)) and f.endswith(".obj")
    ]

    # setup basic (geo) obj reader and writer
    obj_writer = GeoObjWriter()
    reader = GeoObjReader()

    # setup remaining writers
    writers: Dict[str, BaseWriter] = dict()
    writers["cityjson"] = CityJsonWriter()
    writers["geojson"] = GeoJsonWriter()
    writers["geoobj"] = obj_writer
    writers["geooff"] = GeoOffWriter()
    writers["geoply"] = GeoPlyWriter()
    writers["geostl"] = GeoStlWriter()
    writers["geovrml"] = GeoVrmlWriter()
    writers["gml"] = GmlWriter()
    writers["kml"] = KmlWriter()

    # get the number of conversions
    done_jobs = 0
    number_of_jobs = 0
    for value in writers.values():
        number_of_jobs += 1
        if value.supports_origin_base():
            number_of_jobs += 1
    number_of_jobs *= len(inputfiles)
    number_of_jobs += len(inputfiles)

    number_of_vertices = dict()
    # normalize features of obj inputs by reading and writing them with framework based implementations
    for file in inputfiles:
        input_file = reader.read(join(path, file))
        for o in input_file.objects:
            for f in o.faces:
                f.texture_coordinates = []
                f.normal_indices = []
        new_vertices = []
        # normalize the vertices to 13 decimal places
        for vertex in input_file.vertices:
            new_vertex = []
            for c in vertex:
                new_vertex.append(round(c, 13))
            new_vertices.append(new_vertex)
        input_file.vertices = new_vertices
        input_file.texture_coordinates = []
        input_file.normals = []
        number_of_vertices[file] = len(input_file.vertices)
        obj_writer.write(join(target_folder, file), input_file, append_file_type=False)
        done_jobs += 1
        logging.info(
            f"Step {done_jobs} of {number_of_jobs} done (normalize obj features of {file})"
        )

    # prepare local to geo-referenced conversion
    crs = "urn:ogc:def:crs:OGC:2:84"
    origin = [14.2842798233032, 48.30284881591775, 279.807006835938]
    converter = LocalConverter()

    # prepare encodings
    encoding = dict()
    encoding["cityjson"] = False
    encoding["geojson"] = False
    encoding["geoobj"] = False
    encoding["geooff"] = False
    encoding["geoply"] = False
    encoding["geostl"] = False
    encoding["geovrml"] = False
    encoding["gml"] = True
    encoding["kml"] = True

    # prepare result dictionary of file sizes
    sizes = dict()
    origin_based_sizes = dict()
    for file in inputfiles:
        sizes_for_file = dict()
        origin_based_sizes_for_file = dict()
        # read normalized input file
        geoobj = reader.read(join(target_folder, file))

        # convert the input to a origin and a non-origin based geo-referenced representation
        converted = converter.from_local(geoobj, crs, origin, False)
        converted_origin_based = converter.from_local(geoobj, crs, origin, True)

        # write the geo-referenced representation using every configured writer
        for key, value in writers.items():
            value: BaseWriter = value
            folder = join(target_folder, key)
            os.makedirs(folder, exist_ok=True)
            target = join(folder, file)
            value.write(target, converted, encoding[key])
            done_jobs += 1
            logging.info(
                f"Step {done_jobs} of {number_of_jobs} done ({file}, {type(value).__name__})"
            )
            sizes_for_file[key] = os.path.getsize(target + value.get_file_type())
            # if writer supports origin-based representation also create this
            if value.supports_origin_base():
                origin_target = join(folder, "origin_" + file)
                value.write(origin_target, converted_origin_based, encoding[key])
                done_jobs += 1
                logging.info(
                    f"Step {done_jobs} of {number_of_jobs} done ({file}, {type(value).__name__}, origin-based)"
                )
                origin_based_sizes_for_file[key] = os.path.getsize(
                    origin_target + value.get_file_type()
                )

        sizes[file] = sizes_for_file
        origin_based_sizes[file] = origin_based_sizes_for_file

    # create result.csv file with file sizes in KB
    with open(join(target_folder, "result.csv"), "w") as csvfile:
        # prepare csv header
        csvfile.write("file,vertices,original")
        for extension, writer in writers.items():
            csvfile.write(f",{extension}")
            if writer.supports_origin_base():
                csvfile.write(f",{extension}-origin")
        csvfile.write("\n")

        # write csv values (file name followed by writer specific sizes
        for key, size in sizes.items():
            origin_size = origin_based_sizes[key]
            csvfile.write(key)
            csvfile.write(",")
            csvfile.write(str(number_of_vertices[key]))
            csvfile.write(",")
            original_file_size = os.path.getsize(join(target_folder, key))
            csvfile.write("{:.2f}".format(original_file_size / 1024.0))
            for file_format in writers:
                csvfile.write(",")
                csvfile.write("{:.2f}".format(size[file_format] / 1024.0))
                if origin_size.get(file_format) is not None:
                    csvfile.write(",")
                    csvfile.write("{:.2f}".format(origin_size[file_format] / 1024.0))
            csvfile.write("\n")



import copy

from geofiles.conversion.origin_converter import OriginConverter
from geofiles.conversion.transformer import Transformer
from geofiles.domain.geo_object_file import GeoObjectFile


class ExtentCalculator:
    """
    Class which allows to calculate the geographical extents of a GeoObjectFile
    """

    @staticmethod
    def update_extent(
        data: GeoObjectFile,
        include_transformation: bool = False,
        geospatial_extent: bool = False,
        bearing_offset: float = 0.0,
    ) -> GeoObjectFile:
        """
        Updates the min and max extent values of this geo-referenced object file, also considering transformation information and origin-based representation
        :param data: Data to be updated
        :param include_transformation: If true transformation information is considered for the extent information
        :param geospatial_extent: Flag if extent should be georeferenced or contain local coordinates
        :param bearing_offset: Used for conversion (required when geospatial_extent == True or include_transformation == True)
        :returns: Updated data
        """
        origin_converter = OriginConverter()
        origin_based = data.is_origin_based()
        copied_data = copy.deepcopy(data)

        if include_transformation:
            # transformation requires origin based data, so convert the data to origin based
            if not origin_based:
                temp = origin_converter.to_origin(
                    copied_data, bearing_offset=bearing_offset
                )
            else:
                temp = copied_data

            transformer = Transformer()
            transformed = transformer.transform(temp, True, True, True, True)
            if origin_based and not geospatial_extent:
                copied_data.min_extent = transformed.min_extent
                copied_data.max_extent = transformed.max_extent
                return copied_data

            origin = origin_converter.from_origin(
                transformed, bearing_offset=bearing_offset, update_extent=True
            )
            data.min_extent = origin.min_extent
            data.max_extent = origin.max_extent
            return data

        if origin_based and geospatial_extent:  # global origin + global extents
            converted = origin_converter.from_origin(data, bearing_offset, True)
            data.min_extent = converted.min_extent
            data.max_extent = converted.max_extent
        else:  # global coordinates + global extents
            data.update_extent()

        return data



import pathlib

import pkg_resources
from setuptools import setup

try:
    with pathlib.Path('requirements.txt').open() as requirements_txt:
        install_requires = [
            str(requirement)
            for requirement
            in pkg_resources.parse_requirements(requirements_txt)
        ]
except FileNotFoundError:
    install_requires = []

setup(
    package_dir={"": "src"},
    install_requires=install_requires,
)



from abc import ABC
from typing import Iterable

from geofiles.domain.face import Face
from geofiles.domain.geo_object import GeoObject
from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.reader.base import BaseReader


class GeoObjReader(BaseReader, ABC):
    """
    Reader implementaiton for geo-referenced .obj files
    """

    def _read(self, file: Iterable[str]) -> GeoObjectFile:
        """
        Reads a given .geoobj file
        :param file: to be read (may be a string representing the path or an opened file instance)
        :return: Domain representation of the GeoObject
        """
        res = GeoObjectFile()

        current_object = GeoObject()
        last_added_object = None
        found_group = False
        filled_group = False

        for line in file:
            trimmed = line.strip()
            if not trimmed:
                continue
            trimmed = " ".join(trimmed.split())
            # check if current line is a classic vertex
            if trimmed.startswith("v "):
                coordinates = list([float(a) for a in trimmed[2:].split(" ")])
                res.vertices.append(coordinates)
            # check if current line is a face defintion
            elif trimmed.startswith("f "):
                face_defs = trimmed[2:].split(" ")
                face = Face()
                for face_def in face_defs:
                    vals = face_def.split("/")
                    list_len = len(vals)
                    if list_len > 0:
                        face.indices.append(int(vals[0]))
                    if list_len > 1 and vals[1] is not None and len(vals[1]) != 0:
                        face.texture_coordinates.append(int(vals[1]))
                    if list_len > 2:
                        face.normal_indices.append(int(vals[2]))
                current_object.faces.append(face)
                filled_group = True
            # check if current line is a group definition
            elif trimmed.startswith("g ") or trimmed.startswith("o "):
                name = trimmed[2:]
                # if it is the first group definition and if we have not found any other definition
                # just set the name of the current group; otherwise it is a new group
                if not found_group and not filled_group:
                    current_object.name = name
                else:
                    new_object = GeoObject()
                    new_object.name = name
                    new_object.parent = current_object
                    res.objects.append(current_object)
                    last_added_object = current_object
                    current_object = new_object
                found_group = True
            # check if the current line defines the coordinate reference system
            elif trimmed.startswith("crs "):
                res.crs = trimmed[4:]
            # check if the current line defines the origin
            elif trimmed.startswith("or "):
                res.origin = [float(a) for a in trimmed[3:].split(" ")]
            # check if the current line defines a file scale
            elif trimmed.startswith("sc "):
                scale = [float(a) for a in trimmed[3:].split(" ")]
                if not found_group:
                    res.scaling = scale
                else:
                    current_object.scaling = scale
            # check if the current line defines a file translation
            elif trimmed.startswith("t "):
                translation = [float(a) for a in trimmed[2:].split(" ")]
                if not found_group:
                    res.translation = translation
                else:
                    current_object.translation = translation
            # check if the current line defines a file rotation
            elif trimmed.startswith("r "):
                rotation = [float(a) for a in trimmed[2:].split(" ")]
                if not found_group:
                    res.rotation = rotation
                else:
                    current_object.rotation = rotation
            elif trimmed.startswith("e "):
                extent = [float(a) for a in trimmed[2:].split(" ")]
                res.min_extent = extent[:3]
                res.max_extent = extent[3:]
            # check if the current line defines a texture coordinate
            elif trimmed.startswith("vt "):
                coordinates = [float(a) for a in trimmed[3:].split(" ")]
                res.texture_coordinates.append(coordinates)
                filled_group = True
            # check if the current line defines a coordinate's normal
            elif trimmed.startswith("vn "):
                coordinates = [float(a) for a in trimmed[3:].split(" ")]
                res.normals.append(coordinates)
                filled_group = True

        if last_added_object is not current_object:
            res.objects.append(current_object)

        return res



import copy
from typing import Any, List

import pyproj

from geofiles.conversion.static import get_epsg_4326, get_wgs_84, update_min_max
from geofiles.domain.geo_object_file import GeoObjectFile


class CrsConverter:
    """
    Converter for projecting a given geo referenced object to another coordinate system
    """

    def convert(
        self,
        data: GeoObjectFile,
        target_crs: str,
        alwaysxy: bool = True,
        update_extent: bool = False,
    ) -> GeoObjectFile:
        """
        Converts the given data to another coordinate reference system
        :param data: to be converted
        :param target_crs: target coordinate system (string to create a pyproj.crs.CRS)
        :param alwaysxy: If true, the transform method will accept as input and return as output
            coordinates using the traditional GIS order, that is longitude, latitude
            for geographic CRS and easting, northing for most projected CRS.
            Default is false.
        :param update_extent: If true, extent information of the converted data is determined
        :return:
        """
        if data.crs is None:
            raise Exception("Given file is not geo-referenced")

        res = copy.deepcopy(data)
        res.crs = target_crs
        source = data.crs
        from_wgs84 = False
        if source == get_wgs_84():
            source = get_epsg_4326()
            from_wgs84 = True

        target = target_crs
        to_wgs84 = False
        if target_crs == get_wgs_84():
            target = get_epsg_4326()
            to_wgs84 = True
        transformer = pyproj.Transformer.from_crs(source, target, always_xy=alwaysxy)

        if data.is_origin_based() and data.origin is not None:
            res.origin = self._convert_coordinate(
                data.origin, transformer, from_wgs84, to_wgs84
            )

            if update_extent:
                res.update_extent()
        else:
            converted = []
            min_extent: List[float] = []
            max_extent: List[float] = []
            is_first = True
            for vertex in data.vertices:
                converted_vertex = self._convert_coordinate(
                    vertex, transformer, from_wgs84, to_wgs84
                )
                if is_first:
                    min_extent = list(converted_vertex).copy()
                    max_extent = list(converted_vertex).copy()
                    is_first = False
                else:
                    update_min_max(vertex, min_extent, max_extent)
                converted.append(converted_vertex)
            res.vertices = converted
        return res

    @staticmethod
    def _convert_coordinate(
        vertex: List[float], transformer: Any, from_wgs84: bool, to_wgs84: bool
    ) -> List[float]:
        """
        Converts the given vertex based on the given transformer with reference if input or output system is WGS84
        :param vertex: to be converted
        :param transformer: used to transform
        :param from_wgs84: is input wgs84 based?
        :param to_wgs84: is output wgs84 based?
        :return: converted coordinate
        """
        # if source is wgs we have to swap x and y to match EPSG:4326
        if not from_wgs84:
            x = vertex[0]
            y = vertex[1]
        else:
            x = vertex[1]
            y = vertex[0]

        transformed: List[float]
        if len(vertex) > 2:
            z = vertex[2]
            transformed = transformer.transform(x, y, z)
        else:
            transformed = transformer.transform(x, y)

        # if target is wgs we have to swap x and y to convert from EPSG:4326
        if to_wgs84:
            x = transformed[0]
            y = transformed[1]
            z = transformed[2]
            transformed = [y, x, z]
        return transformed



import copy
from typing import Any, List, Optional

from pyproj import Geod

from geofiles.conversion.calculation import (
    get_angle_between_points,
    get_center,
    get_distant_point,
    get_point_distance,
)
from geofiles.conversion.static import (
    get_epsg_4326,
    get_lon_lat,
    get_wgs_84,
    update_min_max,
)
from geofiles.domain.geo_object_file import GeoObjectFile


class OriginConverter:
    """
    Converter used to convert a geo referenced file to a origin based representation or vice versa
    """

    @staticmethod
    def to_origin(
        data: GeoObjectFile,
        origin: Optional[List[Any]] = None,
        bearing_offset: float = 0.0,
        update_extent: bool = False,
    ) -> GeoObjectFile:
        """
        Converts the given file to an origin based representation
        :param data: to be converted
        :param origin: coordinate used as file origin
        :param bearing_offset: angle offset between the origin's coordinate system and the local vertices' coordinate system
        :param update_extent: If true, extent information of the converted data is determined
        :return: origin based representation
        """
        if data.is_origin_based():
            raise Exception("Given geo-referenced object file is already origin based")

        if data.crs != get_wgs_84() and data.crs != get_epsg_4326():
            raise Exception(
                f'Function only supported for "{get_wgs_84()}" and "{get_epsg_4326()}"'
            )

        res = copy.deepcopy(data)
        if origin is None:
            localorigin = get_center(data.vertices)
        else:
            localorigin = origin

        res.origin = localorigin

        new_vertices = []
        geod = Geod(ellps="WGS84")
        min_extent: Optional[List[float]] = data.min_extent
        max_extent: Optional[List[float]] = data.max_extent
        is_first = True
        for vertex in data.vertices:
            lon0, lat0 = get_lon_lat(vertex, data.crs)
            lon1, lat1 = get_lon_lat(localorigin, data.crs)
            bearing, _, distance = geod.inv(lon0, lat0, lon1, lat1)
            new_vertex = get_distant_point(0, 0, distance, bearing + bearing_offset)
            new_vertex.append(localorigin[2] - vertex[2])
            if update_extent:
                if is_first:
                    min_extent = new_vertex.copy()
                    max_extent = new_vertex.copy()
                    is_first = False
                else:
                    update_min_max(new_vertex, min_extent, max_extent)
            new_vertices.append(new_vertex)

        res.vertices = new_vertices
        res.min_extent = min_extent
        res.max_extent = max_extent

        return res

    @staticmethod
    def from_origin(
        data: GeoObjectFile, bearing_offset: float = 0.0, update_extent: bool = False
    ) -> GeoObjectFile:
        """
        Converts the given file from an origin based representation
        :param data: to be converted
        :param bearing_offset: angle offset between the origin's coordinate system and the local vertices' coordinate system
        :param update_extent: If true, extent information of the converted data is determined
        :return: non-origin based representation
        """
        if not data.is_origin_based() or data.origin is None:
            raise Exception("Given geo-referenced object file is not origin based")

        if data.crs != get_wgs_84() and data.crs != get_epsg_4326():
            raise Exception(
                f'Function only supported for "{get_wgs_84()}" and "{get_epsg_4326()}"'
            )

        res = copy.deepcopy(data)
        res.origin = None
        origin_lon, origin_lat = get_lon_lat(data.origin, data.crs)

        new_vertices = []
        geod = Geod(ellps="WGS84")
        origin = [0, 0, 0]
        north_vector = [0, 1]
        if data.vertices:
            min_extent: Optional[List[float]] = data.min_extent
            max_extent: Optional[List[float]] = data.max_extent
            is_first = True
            for vertex in data.vertices:
                distance = get_point_distance(origin, vertex)
                r = get_angle_between_points(vertex, north_vector) + bearing_offset
                new_vertex = geod.fwd(origin_lon, origin_lat, r, distance)
                alt = float(data.origin[2]) - float(vertex[2])
                converted: List[float]
                if data.crs == get_wgs_84():
                    converted = [new_vertex[0], new_vertex[1], alt]
                else:
                    converted = [new_vertex[1], new_vertex[0], alt]
                new_vertices.append(converted)

                if update_extent:
                    if is_first:
                        min_extent = converted.copy()
                        max_extent = converted.copy()
                        is_first = False
                    else:
                        update_min_max(converted, min_extent, max_extent)
            res.min_extent = min_extent
            res.max_extent = max_extent

        res.vertices = new_vertices

        return res



from abc import ABC
from io import TextIOWrapper
from typing import Any

from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.writer.base import BaseWriter


class GeoPlyWriter(BaseWriter, ABC):
    """
    Writer implementation for creating Geo PLY geometry files (.geoply)
    """

    def _write(
        self,
        file: TextIOWrapper,
        data: GeoObjectFile,
        write_binary: bool,
        random_seed: Any,
    ) -> None:
        """
        Write implementation
        :param file: target to be written
        :param data: content to be written
        :param write_binary: flag if file is a binary file
        :return:
        """
        if len(data.objects) != 1:
            raise Exception("GeoPLY can represent only one object. Minimize the data.")

        num_vertices = len(data.vertices)
        faces = []
        for obj in data.objects:
            if (
                obj.contains_scaling()
                or obj.contains_rotation()
                or obj.contains_translation()
            ):
                raise Exception(
                    "GeoPLY does not support local object transformation information."
                )
            for face in obj.faces:
                faces.append(face)
        num_faces = len(faces)
        if data.is_geo_referenced():
            self._write_to_file(file, "geoply", write_binary, True)
        else:
            self._write_to_file(file, "ply", write_binary, True)

        self._write_to_file(file, "format ascii 1.0", write_binary, True)
        if data.is_geo_referenced() and data.crs is not None:
            self._write_to_file(file, f"crs {data.crs}", write_binary, True)

        if data.is_geo_referenced() and data.origin is not None:
            self._write_to_file(
                file,
                f"origin {' '.join([str(f) for f in data.origin])}",
                write_binary,
                True,
            )

        if data.is_geo_referenced() and data.scaling is not None:
            self._write_to_file(
                file,
                f"scale {' '.join([str(f) for f in data.scaling])}",
                write_binary,
                True,
            )

        if data.is_geo_referenced() and data.rotation is not None:
            self._write_to_file(
                file,
                f"rotate {' '.join([str(f) for f in data.rotation])}",
                write_binary,
                True,
            )

        if data.is_geo_referenced() and data.translation is not None:
            self._write_to_file(
                file,
                f"translate {' '.join([str(f) for f in data.translation])}",
                write_binary,
                True,
            )

        if (
            data.contains_extent()
            and data.min_extent is not None
            and data.max_extent is not None
        ):
            self._write_to_file(
                file,
                "extent "
                + " ".join([str(a) for a in data.min_extent])
                + " "
                + " ".join([str(a) for a in data.max_extent]),
                write_binary,
                True,
            )

        self._write_to_file(file, f"element vertex {num_vertices}", write_binary, True)
        self._write_to_file(file, "property float x", write_binary, True)
        self._write_to_file(file, "property float y", write_binary, True)
        self._write_to_file(file, "property float z", write_binary, True)
        self._write_to_file(file, f"element face  {num_faces}", write_binary, True)
        self._write_to_file(
            file, "property list uchar int vertex_index", write_binary, True
        )
        self._write_to_file(file, "end_header", write_binary, True)

        for v in data.vertices:
            self._write_to_file(file, " ".join([str(f) for f in v]), write_binary, True)
        for face in faces:
            s = ""
            for f in face.indices:
                s += " "
                f = int(f)
                if f > 0:
                    s += str(f - 1)
                else:
                    s += str(num_vertices - f)
            self._write_to_file(file, f"{len(face.indices)}{s}", write_binary, True)

    def get_file_type(self) -> str:
        """
        :return: the supported file type of this writer
        """
        return ".geoply"

    def supports_origin_base(self) -> bool:
        """
        :return: true if file format supports origin based representation
        """
        return True



from abc import ABC
from typing import Iterable

from geofiles.domain.face import Face
from geofiles.domain.geo_object import GeoObject
from geofiles.domain.geo_object_file import GeoObjectFile
from geofiles.reader.base import BaseReader


class GeoPlyReader(BaseReader, ABC):
    """
    Reader implementaiton for geo-referenced .ply files (.geoply)
    """

    def _read(self, file: Iterable[str]) -> GeoObjectFile:
        res = GeoObjectFile()
        obj = GeoObject()
        res.objects.append(obj)
        num_of_vertices = 0
        search_for_vertices = False
        cnt = 0

        for line in file:
            trimmed = line.strip()
            if not trimmed:
                continue
            trimmed = " ".join(trimmed.split())
            if not search_for_vertices:
                if trimmed.startswith("crs"):
                    res.crs = trimmed[4:]
                elif trimmed.startswith("element vertex"):
                    num_of_vertices = int(trimmed[14:])
                elif trimmed.startswith("origin"):
                    res.origin = [float(a) for a in trimmed[7:].split(" ")]
                elif trimmed.startswith("scale"):
                    res.scaling = [float(a) for a in trimmed[6:].split(" ")]
                elif trimmed.startswith("translate"):
                    res.translation = [float(a) for a in trimmed[10:].split(" ")]
                elif trimmed.startswith("rotate"):
                    res.rotation = [float(a) for a in trimmed[7:].split(" ")]
                elif trimmed.startswith("extent"):
                    extent = [float(a) for a in trimmed[7:].split(" ")]
                    res.min_extent = extent[:3]
                    res.max_extent = extent[3:]
                elif trimmed.startswith("end_header"):
                    search_for_vertices = True
            else:
                splits = trimmed.split(" ")
                if cnt < num_of_vertices:
                    res.vertices.append([float(a) for a in splits])
                    cnt += 1
                else:
                    face = Face()
                    face.indices = [int(a) + 1 for a in splits[1:]]
                    obj.faces.append(face)

        return res



from typing import List, Optional

from geofiles.conversion.static import is_not_none_nor_empty
from geofiles.domain.face import Face


class GeoObject:
    """
    Class that represents a geo-referenced object
    """

    def __init__(self) -> None:
        """
        A geo-referenced object with the following attributes:
        - name: the name of the individual object
        - parent: reference to a parent object to represent a hierarchy
        - faces: a list of face-objects representing the geometry of this object
        - translation: tuple containing the local translation of origin-based geo objects
        - rotation: tuple containing the local rotation of origin-based geo objects
        - scaling: tuple containing the local scaling of origin-based geo objects
        """
        self.name: str = ""
        self.parent: Optional[GeoObject] = None
        self.faces: List[Face] = []
        self.translation: Optional[List[float]] = None
        self.rotation: Optional[List[float]] = None
        self.scaling: Optional[List[float]] = None

    def contains_scaling(self) -> bool:
        """
        Checks if this geo-referenced file contains global scaling information
        """
        return is_not_none_nor_empty(self.scaling) and self.scaling != [1.0, 1.0, 1.0]

    def contains_translation(self) -> bool:
        """
        Checks if this geo-referenced file contains global translation information
        """
        return is_not_none_nor_empty(self.translation) and self.translation != [
            0.0,
            0.0,
            0.0,
        ]

    def contains_rotation(self) -> bool:
        """
        Checks if this geo-referenced file contains global rotation information
        """
        return is_not_none_nor_empty(self.rotation) and self.rotation != [0.0, 0.0, 0.0]
